#! python

# #####################################################################
# Copyright (c) 2017, Stanley Engle
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions 
# are met:
# 
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in 
#    the documentation and/or other materials provided with the 
#    distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY 
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# #####################################################################

# Standard library imports.
import argparse
import datetime
import math
import sys


# The version number of this script/module.
__version__ = '3'


def _Rs_div_Rso(Rs, Rso):

    rs_div_rso = Rs / Rso

    if rs_div_rso < 0.3:
        rs_div_rso = 0.3

    elif rs_div_rso > 1.0:
        rs_div_rso = 1

    return rs_div_rso


def _calculate(date, lat, ele, Tx, Tn, RHx, RHn, U2, Rs):

    # Step 1.  Extraterrestrial radiation (Ra) is calculated for each  
    # day of each month using equations from Duffie and Beckman (1980).
    Gsc = 0.082
    sigma = 0.00000000490
    G = 0.0

    phi = lat * math.pi / 180.0
    day = int(date.strftime('%j'))
    angle = 2 * math.pi / 365 * day
    delta = 0.409 * math.sin(angle -1.39)
    omegas = math.acos(-1.00 * math.tan(phi) * math.tan(delta))
    dr = 1 + 0.033 * math.cos(angle)

    s1_term1 = (24.0 * 60.0) / math.pi
    s1_term2 = omegas * math.sin(delta) * math.sin(phi)
    s1_term3 = math.cos(phi) * math.cos(delta) * math.sin(omegas)

    Ra = s1_term1 * Gsc * dr * (s1_term2 + s1_term3)

    # STEP 2.  Calculate the daily net radiation (Rn) expected over 
    # grass in MJ/m^2/d using equations from Allen et al.(1994). 
    Rso = Ra * (0.75 + 0.00002 * ele)
    Rns = (1 - 0.23) * Rs
    f = (1.35 * (_Rs_div_Rso(Rs, Rso))) - 0.35

    e_sub_s = lambda T: 0.6108 * math.exp((17.27 * T) / (T + 237.3))
    esTx = e_sub_s(Tx)
    esTn = e_sub_s(Tn)

    ed = (esTx * (RHn / 100.0) + esTn * (RHx / 100.0)) / 2.0
    epsilon_prime = 0.34 - 0.14 * math.sqrt(ed)

    lf = lambda T: math.pow(T + 273.15, 4)
    s2_term1 = -1.0 * f * epsilon_prime * sigma
    s2_term2 = (lf(Tx) + lf(Tn)) / 2.0
    Rnl = s2_term1 * s2_term2

    Rn = Rns + Rnl

    # Step 3.  Calculate variables needed to compute ETh, ETo and 
    # ETr.
    s3_term = math.pow((293 - 0.0065 * ele) / 293.0, 5.26)
    beta = 101.3 * s3_term
    gamma = 0.00163 * (beta / 2.45)
    Tm = (Tx + Tn) / 2.0
    e_super_o = e_sub_s(Tm)
    Delta = (4099 * e_super_o) / math.pow(Tm + 237.3, 2)
    ea = (esTx + esTn) / 2.0

    # Step 4.  Calculate ETh using the Hargreaves-Samani (1982) 
    # equation.
    s4_term1 = 0.0023 * Ra
    s4_term2 = ((Tx + Tn) / 2.0) + 17.8
    s4_term3 = math.sqrt(Tx - Tn)
    ETh = 0.408 * (s4_term1 * s4_term2 * s4_term3)

    # Step 5.  Calculate ETo using the ASCE-EWRI (2004) standardized 
    # equation for short canopy reference ET.
    s5_term1 = 0.408 * Delta * (Rn - G)
    s5_term2 = Delta + gamma * (1 + 0.34 * U2)
    Ro = s5_term1 / s5_term2

    s5_term3 = (900 * gamma) / (Tm + 273)
    s5_term4 = (ea - ed)
    Ao = (s5_term3 * U2 * s5_term4) / s5_term2

    ETo = Ro + Ao

    # Step 6.  Calculate ETr using the ASCE-EWRI (2004) standardized 
    # equation for tall canopy reference ET.
    s6_term1 = 0.408 * Delta * (Rn - G)
    s6_term2 = Delta + gamma * (1 + 0.38 * U2)
    Rr = s6_term1 / s6_term2

    s6_term3 = (1600 * gamma) / (Tm + 273)
    s6_term4 = (ea - ed)
    Ar = (s6_term3 * U2 * s6_term4) / s6_term2

    ETr = Rr + Ar

    return (ETh, ETo, ETr)


def calculate(date, lat, ele, Tx, Tn, RHx, RHn, U2, Rs, 
    on_error = float('nan')):
    """Calculate the Reference ET values.

    This function calculates Hargreaves-Samani (ETh) and ASCE-EWRI 
    (ETo, ETr) reference ET values.

    Parameters
    ----------
    date : datetime.date
        The date when the meteorological input values were collected.
    lat : float
        The latitude where the meteorological input values were 
        collected.
    ele : float
        The elevation, in meters, where the meteorological input 
        values were collected.
    Tx : float
        The maximum daily air temperature in Celsius.
    Tn : float
        The minimum daily air temperature in Celsius.
    RHx : float
        The maximum daily relative humidity.
    RHn : float
        The minimum daily relative humidity.
    U2 : float
        The mean daily wind speed, in meters per second.
    Rs : float
        The total daily solar radiation, in MJ/m^2.
    on_error : optional, float('nan')
        The return values when an error occcurs during calculation.  
        May be any type, but preferably a float.

    Returns
    -------
    tuple
        A tuple containing float values of the ETh, ETo, and ETr 
        values.

    """
    try:

        _date = valid_date(str(date))
        _lat = float(lat)
        _ele = float(ele)
        _Tx = float(Tx)
        _Tn = float(Tn)
        _RHx = float(RHx)
        _RHn = float(RHn)
        _U2 = float(U2)
        _Rs = float(Rs)

        etref = \
            _calculate(_date, _lat, _ele, _Tx, _Tn, _RHx, _RHn, _U2, _Rs)

    except:
        etref = (on_error, on_error, on_error)

    return etref


def run(date, lat, ele, Tx, Tn, RHx, RHn, U2, Rs, on_error = float('nan')):
    """Calls the calculate function.  See calculate docstring.

    """
    c = calculate

    return c(date, lat, ele, Tx, Tn, RHx, RHn, U2, Rs, on_error = on_error)

# This function is used to determine the validity of the date arguments 
# passed via the command line.
def valid_date(date_str):
    return datetime.datetime.strptime(date_str, '%Y-%m-%d').date()


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('date', 
        type = valid_date, 
        help = 'The date, in the format YYYY-MM-DD.')
    parser.add_argument('lat', 
        type = float, 
        help = 'The latitude, in decimal degrees.')
    parser.add_argument('ele', 
        type = float, 
        help = 'The elevation, in meters.')
    parser.add_argument('Tx', 
        type = float, 
        help = 'The maximum daily air temperature in degrees Celsius.')
    parser.add_argument('Tn', 
        type = float, 
        help = 'The minimum daily air temperature in degrees Celsius.')
    parser.add_argument('RHx', 
        type = float, 
        help = 'The maximum daily RH value.')
    parser.add_argument('RHn', 
        type = float, 
        help = 'The minimum daily RH value.')
    parser.add_argument('U2', 
        type = float, 
        help = 'The mean daily wind speed, in meters per second.')
    parser.add_argument('Rs', 
        type = float, 
        help = 'The total daily solar radiation, in MJ/m^2.')

    # Parse the command line arguments
    args = parser.parse_args()

    # Call the run function.
    (ETh, ETo, ETr) = run(args.date, 
        args.lat, 
        args.ele, 
        args.Tx, 
        args.Tn, 
        args.RHx, 
        args.RHn, 
        args.U2, 
        args.Rs)

    # Print the output to standard out.
    sys.stdout.write(str(ETh) + ', ' + str(ETo) + ', ' + str(ETr) + '\n')

    # Explicitly exit the script.
    sys.exit(0)


